@@ -59,8 +59,8 @@ struct User {
         GKeyFile     *keyfile;
 
         gid_t         gid;
-        gint64        expiration_time;
-        gint64        last_change_time;
+        GDateTime    *user_expiration_time;
+        GDateTime    *last_change_time;
         gint64        min_days_between_changes;
         gint64        max_days_between_changes;
         gint64        days_to_warn;
@@ -68,8 +68,10 @@ struct User {
         GVariant     *login_history;
         gchar        *icon_file;
         gchar        *default_icon_file;
+        gchar        *gecos;
         gboolean      account_expiration_policy_known;
         gboolean      cached;
+        gboolean      template_loaded;
 
         guint        *extension_ids;
         guint         n_extension_ids;
@@ -83,6 +85,7 @@ typedef struct UserClass
 } UserClass;
 
 static void user_accounts_user_iface_init (AccountsUserIface *iface);
+static void user_update_from_keyfile (User *user, GKeyFile *keyfile);
 
 G_DEFINE_TYPE_WITH_CODE (User, user, ACCOUNTS_TYPE_USER_SKELETON, G_IMPLEMENT_INTERFACE (ACCOUNTS_TYPE_USER, user_accounts_user_iface_init));
 
@@ -137,6 +140,261 @@ user_reset_icon_file (User *user)
         }
 }
 
+static gboolean
+user_has_cache_file (User *user)
+{
+        g_autofree char *filename = NULL;
+
+        filename = g_build_filename (USERDIR, user_get_user_name (user), NULL);
+
+        return g_file_test (filename, G_FILE_TEST_EXISTS);
+}
+
+static gboolean
+is_valid_shell_identifier_character (char     c,
+                                     gboolean first)
+{
+        return (!first && g_ascii_isdigit (c)) ||
+                c == '_' ||
+                g_ascii_isalpha (c);
+}
+
+static char *
+expand_template_variables (User       *user,
+                           GHashTable *template_variables,
+                           const char *str)
+{
+        GString *s = g_string_new ("");
+        const char *p, *start;
+        char c;
+
+        p = str;
+        while (*p) {
+                c = *p;
+                if (c == '\\') {
+                        p++;
+                        c = *p;
+                        if (c != '\0') {
+                                p++;
+                                switch (c) {
+                                case '\\':
+                                        g_string_append_c (s, '\\');
+                                        break;
+                                case '$':
+                                        g_string_append_c (s, '$');
+                                        break;
+                                default:
+                                        g_string_append_c (s, '\\');
+                                        g_string_append_c (s, c);
+                                        break;
+                                }
+                        }
+                } else if (c == '$') {
+                        gboolean brackets = FALSE;
+                        p++;
+                        if (*p == '{') {
+                                brackets = TRUE;
+                                p++;
+                        }
+                        start = p;
+                        while (*p != '\0' &&
+                               is_valid_shell_identifier_character (*p, p == start))
+                                p++;
+                        if (p == start || (brackets && *p != '}')) {
+                                g_string_append_c (s, '$');
+                                if (brackets)
+                                        g_string_append_c (s, '{');
+                                g_string_append_len (s, start, p - start);
+                        } else {
+                                g_autofree char *variable = NULL;
+                                const char *value;
+
+                                if (brackets && *p == '}')
+                                        p++;
+
+                                variable = g_strndup (start, p - start - 1);
+
+                                value = g_hash_table_lookup (template_variables, variable);
+                                if (value) {
+                                        g_string_append (s, value);
+                                }
+                        }
+                } else {
+                        p++;
+                        g_string_append_c (s, c);
+                }
+        }
+        return g_string_free (s, FALSE);
+}
+
+static void
+load_template_environment_file (User       *user,
+                                GHashTable *variables,
+                                const char *file)
+{
+        g_autofree char *contents = NULL;
+        g_auto (GStrv) lines = NULL;
+        g_autoptr (GError) error = NULL;
+        gboolean file_loaded;
+        size_t i;
+
+        file_loaded = g_file_get_contents (file, &contents, NULL, &error);
+
+        if (!file_loaded) {
+                g_debug ("Couldn't load template environment file %s: %s",
+                         file, error->message);
+                return;
+        }
+
+        lines = g_strsplit (contents, "\n", -1);
+
+        for (i = 0; lines[i] != NULL; i++) {
+                char *p;
+                char *variable_end;
+                const char *variable;
+                const char *value;
+
+                p = lines[i];
+                while (g_ascii_isspace (*p))
+                        p++;
+                if (*p == '#' || *p == '\0')
+                        continue;
+                variable = p;
+                while (is_valid_shell_identifier_character (*p, p == variable))
+                        p++;
+                variable_end = p;
+                while (g_ascii_isspace (*p))
+                        p++;
+                if (variable_end == variable || *p != '=') {
+                        g_debug ("template environment file %s has invalid line '%s'\n", file, lines[i]);
+                        continue;
+                }
+                *variable_end = '\0';
+                p++;
+                while (g_ascii_isspace (*p))
+                        p++;
+                value = p;
+
+                if (g_hash_table_lookup (variables, variable) == NULL) {
+                        g_hash_table_insert (variables,
+                                             g_strdup (variable),
+                                             g_strdup (value));
+                }
+
+        }
+}
+
+static void
+initialize_template_environment (User               *user,
+                                 GHashTable         *variables,
+                                 const char * const *files)
+{
+        size_t i;
+
+        g_hash_table_insert (variables, g_strdup ("HOME"), g_strdup (accounts_user_get_home_directory (ACCOUNTS_USER (user))));
+        g_hash_table_insert (variables, g_strdup ("USER"), g_strdup (user_get_user_name (user)));
+
+        if (files == NULL)
+                return;
+
+        for (i = 0; files[i] != NULL; i++) {
+                load_template_environment_file (user, variables, files[i]);
+        }
+}
+
+static void
+user_update_from_template (User *user)
+{
+        g_autofree char *filename = NULL;
+        g_autoptr (GKeyFile) key_file = NULL;
+        g_autoptr (GError) error = NULL;
+        g_autoptr (GHashTable) template_variables = NULL;
+        g_auto (GStrv) template_environment_files = NULL;
+        gboolean key_file_loaded = FALSE;
+        const char * const *system_dirs[] = {
+                (const char *[]) { "/run", SYSCONFDIR, NULL },
+                g_get_system_data_dirs (),
+                NULL
+        };
+        g_autoptr (GPtrArray) dirs = NULL;
+        AccountType account_type;
+        const char *account_type_string;
+        size_t i, j;
+        g_autofree char *contents = NULL;
+        g_autofree char *expanded = NULL;
+        g_auto (GStrv) lines = NULL;
+
+        if (user->template_loaded)
+                return;
+
+        filename = g_build_filename (USERDIR,
+                                     accounts_user_get_user_name (ACCOUNTS_USER (user)),
+                                     NULL);
+
+        account_type = accounts_user_get_account_type (ACCOUNTS_USER (user));
+        if (account_type == ACCOUNT_TYPE_ADMINISTRATOR)
+                account_type_string = "administrator";
+        else
+                account_type_string = "standard";
+
+        dirs = g_ptr_array_new ();
+        for (i = 0; system_dirs[i] != NULL; i++) {
+                for (j = 0; system_dirs[i][j] != NULL; j++) {
+                        char *dir;
+
+                        dir = g_build_filename (system_dirs[i][j],
+                                                "accountsservice",
+                                                "user-templates",
+                                                NULL);
+                        g_ptr_array_add (dirs, dir);
+                }
+        }
+        g_ptr_array_add (dirs, NULL);
+
+        key_file = g_key_file_new ();
+        key_file_loaded = g_key_file_load_from_dirs (key_file,
+                                                     account_type_string,
+                                                     (const char **) dirs->pdata,
+                                                     NULL,
+                                                     G_KEY_FILE_KEEP_COMMENTS,
+                                                     &error);
+
+        if (!key_file_loaded) {
+                g_debug ("failed to load user template: %s", error->message);
+                return;
+        }
+
+        template_variables = g_hash_table_new_full (g_str_hash,
+                                                    g_str_equal,
+                                                    g_free,
+                                                    g_free);
+
+        template_environment_files = g_key_file_get_string_list (key_file,
+                                                                 "Template",
+                                                                 "EnvironmentFiles",
+                                                                 NULL,
+                                                                 NULL);
+
+        initialize_template_environment (user, template_variables, (const char * const *) template_environment_files);
+
+        g_key_file_remove_group (key_file, "Template", NULL);
+        contents = g_key_file_to_data (key_file, NULL, NULL);
+        lines = g_strsplit (contents, "\n", -1);
+
+        expanded = expand_template_variables (user, template_variables, contents);
+
+        key_file_loaded = g_key_file_load_from_data (key_file,
+                                                     expanded,
+                                                     strlen (expanded),
+                                                     G_KEY_FILE_KEEP_COMMENTS,
+                                                     &error);
+
+        if (key_file_loaded)
+                user_update_from_keyfile (user, key_file);
+
+        user->template_loaded = key_file_loaded;
+}
+
 void
 user_update_from_pwent (User          *user,
                         struct passwd *pwent,
@@ -145,12 +403,14 @@ user_update_from_pwent (User          *user,
         g_autofree gchar *real_name = NULL;
         gboolean is_system_account;
         const gchar *passwd;
+        g_autoptr(GDateTime) start_time = NULL;
         gboolean locked;
         PasswordMode mode;
         AccountType account_type;
 
         g_object_freeze_notify (G_OBJECT (user));
 
+        g_clear_pointer (&user->gecos, g_free);
         if (pwent->pw_gecos && pwent->pw_gecos[0] != '\0') {
                 gchar *first_comma = NULL;
                 gchar *valid_utf8_name = NULL;
@@ -158,6 +418,7 @@ user_update_from_pwent (User          *user,
                 if (g_utf8_validate (pwent->pw_gecos, -1, NULL)) {
                         valid_utf8_name = pwent->pw_gecos;
                         first_comma = g_utf8_strchr (valid_utf8_name, -1, ',');
+                        user->gecos = g_strdup (pwent->pw_gecos);
                 }
                 else {
                         g_warning ("User %s has invalid UTF-8 in GECOS field. "
@@ -223,8 +484,11 @@ user_update_from_pwent (User          *user,
                         mode = PASSWORD_MODE_SET_AT_LOGIN;
                 }
 
-                user->expiration_time = spent->sp_expire;
-                user->last_change_time  = spent->sp_lstchg;
+                start_time = g_date_time_new_from_unix_utc (0);
+
+                user->user_expiration_time = g_date_time_add_days (start_time, spent->sp_expire);
+                user->last_change_time = g_date_time_add_days (start_time, spent->sp_lstchg);
+
                 user->min_days_between_changes = spent->sp_min;
                 user->max_days_between_changes = spent->sp_max;
                 user->days_to_warn  = spent->sp_warn;
@@ -239,17 +503,17 @@ user_update_from_pwent (User          *user,
                                                      passwd);
         accounts_user_set_system_account (ACCOUNTS_USER (user), is_system_account);
 
+        if (!user_has_cache_file (user))
+                user_update_from_template (user);
         g_object_thaw_notify (G_OBJECT (user));
 }
 
-void
+static void
 user_update_from_keyfile (User     *user,
                           GKeyFile *keyfile)
 {
         gchar *s;
 
-        g_object_freeze_notify (G_OBJECT (user));
-
         s = g_key_file_get_string (keyfile, "User", "Language", NULL);
         if (s != NULL) {
                 accounts_user_set_language (ACCOUNTS_USER (user), s);
@@ -310,9 +574,25 @@ user_update_from_keyfile (User     *user,
 
         g_clear_pointer (&user->keyfile, g_key_file_unref);
         user->keyfile = g_key_file_ref (keyfile);
+}
+
+void
+user_update_from_cache (User *user)
+{
+        g_autofree gchar *filename = NULL;
+        g_autoptr(GKeyFile) key_file = NULL;
+
+        filename = g_build_filename (USERDIR, accounts_user_get_user_name (ACCOUNTS_USER (user)), NULL);
+
+        key_file = g_key_file_new ();
+
+        if (!g_key_file_load_from_file (key_file, filename, 0, NULL))
+                return;
+
+        g_object_freeze_notify (G_OBJECT (user));
+        user_update_from_keyfile (user, key_file);
         user_set_cached (user, TRUE);
         user_set_saved (user, TRUE);
-
         g_object_thaw_notify (G_OBJECT (user));
 }
 
@@ -536,6 +816,9 @@ user_extension_authentication_done (Daemon                *daemon,
         GDBusInterfaceInfo *interface = user_data;
         const gchar *method_name;
 
+        if (!user_has_cache_file (user))
+                user_update_from_template (user);
+
         method_name = g_dbus_method_invocation_get_method_name (invocation);
 
         if (g_str_equal (method_name, "Get"))
@@ -611,7 +894,7 @@ user_extension_method_call (GDBusConnection       *connection,
                 user_extension_authentication_done (user->daemon, user, invocation, iface_info);
         }
         else {
-                daemon_local_check_auth (user->daemon, user, action_id, TRUE,
+                daemon_local_check_auth (user->daemon, user, action_id,
                                          user_extension_authentication_done,
                                          invocation, iface_info, NULL);
         }
@@ -834,7 +1117,9 @@ user_change_real_name_authorized_cb (Daemon                *daemon,
 
 {
         gchar *name = data;
+        g_autofree gchar *new_gecos = NULL;
         g_autoptr(GError) error = NULL;
+        const gchar *first_comma = NULL;
         const gchar *argv[6];
 
         if (g_strcmp0 (accounts_user_get_real_name (ACCOUNTS_USER (user)), name) != 0) {
@@ -844,9 +1129,21 @@ user_change_real_name_authorized_cb (Daemon                *daemon,
                          accounts_user_get_uid (ACCOUNTS_USER (user)),
                          name);
 
+                if (user->gecos != NULL)
+                        first_comma = g_utf8_strchr (user->gecos, -1, ',');
+
+                if (first_comma != NULL) {
+                        /* Preserve the existing value of the GECOS
+                         * except for the first element, full name.
+                         */
+                        new_gecos = g_strconcat (name, first_comma, NULL);
+                } else {
+                        new_gecos = g_strdup (name);
+                }
+
                 argv[0] = "/usr/sbin/usermod";
                 argv[1] = "-c";
-                argv[2] = name;
+                argv[2] = new_gecos;
                 argv[3] = "--";
                 argv[4] = accounts_user_get_user_name (ACCOUNTS_USER (user));
                 argv[5] = NULL;
@@ -873,7 +1170,12 @@ user_set_real_name (AccountsUser          *auser,
 
         if (!get_caller_uid (context, &uid)) {
                 throw_error (context, ERROR_FAILED, "identifying caller failed");
-                return FALSE;
+                return TRUE;
+        }
+
+        if (g_utf8_strchr (real_name, -1, ',') != NULL) {
+                throw_error (context, ERROR_FAILED, "setting real name failed: real name '%s' must not contain commas", real_name);
+                return TRUE;
         }
 
         if (accounts_user_get_uid (ACCOUNTS_USER (user)) == (uid_t) uid)
@@ -884,7 +1186,6 @@ user_set_real_name (AccountsUser          *auser,
         daemon_local_check_auth (user->daemon,
                                  user,
                                  action_id,
-                                 TRUE,
                                  user_change_real_name_authorized_cb,
                                  context,
                                  g_strdup (real_name),
@@ -943,7 +1244,6 @@ user_set_user_name (AccountsUser          *auser,
         daemon_local_check_auth (user->daemon,
                                  user,
                                  "org.freedesktop.accounts.user-administration",
-                                 TRUE,
                                  user_change_user_name_authorized_cb,
                                  context,
                                  g_strdup (user_name),
@@ -983,7 +1283,7 @@ user_set_email (AccountsUser          *auser,
 
         if (!get_caller_uid (context, &uid)) {
                 throw_error (context, ERROR_FAILED, "identifying caller failed");
-                return FALSE;
+                return TRUE;
         }
 
         if (accounts_user_get_uid (ACCOUNTS_USER (user)) == (uid_t) uid)
@@ -994,7 +1294,6 @@ user_set_email (AccountsUser          *auser,
         daemon_local_check_auth (user->daemon,
                                  user,
                                  action_id,
-                                 TRUE,
                                  user_change_email_authorized_cb,
                                  context,
                                  g_strdup (email),
@@ -1034,7 +1333,7 @@ user_set_language (AccountsUser          *auser,
 
         if (!get_caller_uid (context, &uid)) {
                 throw_error (context, ERROR_FAILED, "identifying caller failed");
-                return FALSE;
+                return TRUE;
         }
 
         if (accounts_user_get_uid (ACCOUNTS_USER (user)) == (uid_t) uid)
@@ -1045,7 +1344,6 @@ user_set_language (AccountsUser          *auser,
         daemon_local_check_auth (user->daemon,
                                  user,
                                  action_id,
-                                 TRUE,
                                  user_change_language_authorized_cb,
                                  context,
                                  g_strdup (language),
@@ -1083,7 +1381,7 @@ user_set_session (AccountsUser          *auser,
 
         if (!get_caller_uid (context, &uid)) {
                 throw_error (context, ERROR_FAILED, "identifying caller failed");
-                return FALSE;
+                return TRUE;
         }
 
         if (accounts_user_get_uid (ACCOUNTS_USER (user)) == (uid_t) uid)
@@ -1094,7 +1392,6 @@ user_set_session (AccountsUser          *auser,
         daemon_local_check_auth (user->daemon,
                                  user,
                                  action_id,
-                                 TRUE,
                                  user_change_session_authorized_cb,
                                  context,
                                  g_strdup (session),
@@ -1132,7 +1429,7 @@ user_set_session_type (AccountsUser          *auser,
 
         if (!get_caller_uid (context, &uid)) {
                 throw_error (context, ERROR_FAILED, "identifying caller failed");
-                return FALSE;
+                return TRUE;
         }
 
         if (accounts_user_get_uid (ACCOUNTS_USER (user)) == (uid_t) uid)
@@ -1143,7 +1440,6 @@ user_set_session_type (AccountsUser          *auser,
         daemon_local_check_auth (user->daemon,
                                  user,
                                  action_id,
-                                 TRUE,
                                  user_change_session_type_authorized_cb,
                                  context,
                                  g_strdup (session_type),
@@ -1181,7 +1477,7 @@ user_set_x_session (AccountsUser          *auser,
 
         if (!get_caller_uid (context, &uid)) {
                 throw_error (context, ERROR_FAILED, "identifying caller failed");
-                return FALSE;
+                return TRUE;
         }
 
         if (accounts_user_get_uid (ACCOUNTS_USER (user)) == (uid_t) uid)
@@ -1192,7 +1488,6 @@ user_set_x_session (AccountsUser          *auser,
         daemon_local_check_auth (user->daemon,
                                  user,
                                  action_id,
-                                 TRUE,
                                  user_change_x_session_authorized_cb,
                                  context,
                                  g_strdup (x_session),
@@ -1208,15 +1503,20 @@ user_get_password_expiration_policy_authorized_cb (Daemon                *daemon
                                                    gpointer               data)
 
 {
+        guint64 user_expiration_time;
+        guint64 last_change_time;
+
         if (!user->account_expiration_policy_known) {
                 throw_error (context, ERROR_NOT_SUPPORTED, "account expiration policy unknown to accounts service");
                 return;
         }
 
+        user_expiration_time = g_date_time_to_unix (user->user_expiration_time);
+        last_change_time = g_date_time_to_unix (user->last_change_time);
         accounts_user_complete_get_password_expiration_policy (ACCOUNTS_USER (user),
                                                                context,
-                                                               user->expiration_time,
-                                                               user->last_change_time,
+                                                               user_expiration_time,
+                                                               last_change_time,
                                                                user->min_days_between_changes,
                                                                user->max_days_between_changes,
                                                                user->days_to_warn,
@@ -1233,7 +1533,7 @@ user_get_password_expiration_policy (AccountsUser          *auser,
 
         if (!get_caller_uid (context, &uid)) {
                 throw_error (context, ERROR_FAILED, "identifying caller failed");
-                return FALSE;
+                return TRUE;
         }
 
         if (accounts_user_get_uid (ACCOUNTS_USER (user)) == (uid_t) uid)
@@ -1244,7 +1544,6 @@ user_get_password_expiration_policy (AccountsUser          *auser,
         daemon_local_check_auth (user->daemon,
                                  user,
                                  action_id,
-                                 TRUE,
                                  user_get_password_expiration_policy_authorized_cb,
                                  context,
                                  NULL,
@@ -1283,7 +1582,7 @@ user_set_location (AccountsUser          *auser,
 
         if (!get_caller_uid (context, &uid)) {
                 throw_error (context, ERROR_FAILED, "identifying caller failed");
-                return FALSE;
+                return TRUE;
         }
 
         if (accounts_user_get_uid (ACCOUNTS_USER (user)) == (uid_t) uid)
@@ -1294,7 +1593,6 @@ user_set_location (AccountsUser          *auser,
         daemon_local_check_auth (user->daemon,
                                  user,
                                  action_id,
-                                 TRUE,
                                  user_change_location_authorized_cb,
                                  context,
                                  g_strdup (location),
@@ -1351,7 +1649,6 @@ user_set_home_directory (AccountsUser          *auser,
         daemon_local_check_auth (user->daemon,
                                  user,
                                  "org.freedesktop.accounts.user-administration",
-                                 TRUE,
                                  user_change_home_dir_authorized_cb,
                                  context,
                                  g_strdup (home_dir),
@@ -1405,7 +1702,6 @@ user_set_shell (AccountsUser          *auser,
         daemon_local_check_auth (user->daemon,
                                  user,
                                  "org.freedesktop.accounts.user-administration",
-                                 TRUE,
                                  user_change_shell_authorized_cb,
                                  context,
                                  g_strdup (shell),
@@ -1564,7 +1860,7 @@ user_set_icon_file (AccountsUser          *auser,
 
         if (!get_caller_uid (context, &uid)) {
                 throw_error (context, ERROR_FAILED, "identifying caller failed");
-                return FALSE;
+                return TRUE;
         }
 
         if (accounts_user_get_uid (ACCOUNTS_USER (user)) == (uid_t) uid)
@@ -1575,7 +1871,6 @@ user_set_icon_file (AccountsUser          *auser,
         daemon_local_check_auth (user->daemon,
                                  user,
                                  action_id,
-                                 TRUE,
                                  user_change_icon_file_authorized_cb,
                                  context,
                                  g_strdup (filename),
@@ -1653,7 +1948,6 @@ user_set_locked (AccountsUser          *auser,
         daemon_local_check_auth (user->daemon,
                                  user,
                                  "org.freedesktop.accounts.user-administration",
-                                 TRUE,
                                  user_change_locked_authorized_cb,
                                  context,
                                  GINT_TO_POINTER (locked),
@@ -1675,8 +1969,9 @@ user_change_account_type_authorized_cb (Daemon                *daemon,
         gint ngroups;
         g_autoptr(GString) str = NULL;
         g_auto(GStrv) extra_admin_groups = NULL;
+        g_autofree gid_t *extra_admin_groups_gids = NULL;
+        gsize n_extra_admin_groups_gids = 0;
         gid_t admin_gid;
-        struct group *grp;
         gint i;
         const gchar *argv[6];
 
@@ -1687,33 +1982,30 @@ user_change_account_type_authorized_cb (Daemon                *daemon,
                          accounts_user_get_uid (ACCOUNTS_USER (user)),
                          account_type);
 
-                grp = getgrnam (ADMIN_GROUP);
-                if (grp == NULL) {
+                if (!get_admin_groups (&admin_gid, &extra_admin_groups_gids, &n_extra_admin_groups_gids)) {
                         throw_error (context, ERROR_FAILED, "failed to set account type: " ADMIN_GROUP " group not found");
                         return;
                 }
-                admin_gid = grp->gr_gid;
 
                 ngroups = get_user_groups (accounts_user_get_user_name (ACCOUNTS_USER (user)), user->gid, &groups);
 
                 str = g_string_new ("");
                 for (i = 0; i < ngroups; i++) {
+                        gboolean group_is_admin = FALSE;
+
                         if (groups[i] == admin_gid)
-                                continue;
-                        g_string_append_printf (str, "%d,", groups[i]);
+                                group_is_admin = TRUE;
+                        for (gsize j = 0; j < n_extra_admin_groups_gids; j++)
+                                if (groups[i] == extra_admin_groups_gids[j])
+                                        group_is_admin = TRUE;
+
+                        if (!group_is_admin)
+                                g_string_append_printf (str, "%d,", groups[i]);
                 }
                 switch (account_type) {
                 case ACCOUNT_TYPE_ADMINISTRATOR:
-                        extra_admin_groups = g_strsplit (EXTRA_ADMIN_GROUPS, ",", 0);
-
-                        for (i = 0; extra_admin_groups[i] != NULL; i++) {
-                                struct group *extra_group;
-                                extra_group = getgrnam (extra_admin_groups[i]);
-                                if (extra_group == NULL || extra_group->gr_gid == admin_gid)
-                                        continue;
-
-                                g_string_append_printf (str, "%d,", extra_group->gr_gid);
-                        }
+                        for (i = 0; i < n_extra_admin_groups_gids; i++)
+                                g_string_append_printf (str, "%d,", extra_admin_groups_gids[i]);
 
                         g_string_append_printf (str, "%d", admin_gid);
                         break;
@@ -1756,13 +2048,12 @@ user_set_account_type (AccountsUser          *auser,
         User *user = (User*)auser;
         if (account_type < 0 || account_type > ACCOUNT_TYPE_LAST) {
                 throw_error (context, ERROR_FAILED, "unknown account type: %d", account_type);
-                return FALSE;
+                return TRUE;
         }
 
         daemon_local_check_auth (user->daemon,
                                  user,
                                  "org.freedesktop.accounts.user-administration",
-                                 TRUE,
                                  user_change_account_type_authorized_cb,
                                  context,
                                  GINT_TO_POINTER (account_type),
@@ -1862,12 +2153,12 @@ user_set_password_mode (AccountsUser          *auser,
 
         if (mode < 0 || mode > PASSWORD_MODE_LAST) {
                 throw_error (context, ERROR_FAILED, "unknown password mode: %d", mode);
-                return FALSE;
+                return TRUE;
         }
 
         if (!get_caller_uid (context, &uid)) {
                 throw_error (context, ERROR_FAILED, "identifying caller failed");
-                return FALSE;
+                return TRUE;
         }
 
         if (accounts_user_get_uid (ACCOUNTS_USER (user)) == (uid_t) uid)
@@ -1878,7 +2169,6 @@ user_set_password_mode (AccountsUser          *auser,
         daemon_local_check_auth (user->daemon,
                                  user,
                                  action_id,
-                                 TRUE,
                                  user_change_password_mode_authorized_cb,
                                  context,
                                  GINT_TO_POINTER (mode),
@@ -1948,7 +2238,7 @@ user_set_password (AccountsUser          *auser,
 
         if (!get_caller_uid (context, &uid)) {
                 throw_error (context, ERROR_FAILED, "identifying caller failed");
-                return FALSE;
+                return TRUE;
         }
 
         data = g_new (gchar *, 3);
@@ -1964,7 +2254,6 @@ user_set_password (AccountsUser          *auser,
         daemon_local_check_auth (user->daemon,
                                  user,
                                  action_id,
-                                 TRUE,
                                  user_change_password_authorized_cb,
                                  context,
                                  data,
@@ -2008,7 +2297,7 @@ user_set_password_hint (AccountsUser          *auser,
 
         if (!get_caller_uid (context, &uid)) {
                 throw_error (context, ERROR_FAILED, "identifying caller failed");
-                return FALSE;
+                return TRUE;
         }
 
         if (accounts_user_get_uid (ACCOUNTS_USER (user)) == (uid_t) uid)
@@ -2019,7 +2308,6 @@ user_set_password_hint (AccountsUser          *auser,
         daemon_local_check_auth (user->daemon,
                                  user,
                                  action_id,
-                                 TRUE,
                                  user_change_password_hint_authorized_cb,
                                  context,
                                  g_strdup (hint),
@@ -2065,7 +2353,6 @@ user_set_automatic_login (AccountsUser          *auser,
         daemon_local_check_auth (user->daemon,
                                  user,
                                  "org.freedesktop.accounts.user-administration",
-                                 TRUE,
                                  user_change_automatic_login_authorized_cb,
                                  context,
                                  GINT_TO_POINTER (enabled),
@@ -2087,8 +2374,11 @@ user_finalize (GObject *object)
         g_clear_pointer (&user->keyfile, g_key_file_unref);
 
         g_free (user->default_icon_file);
+        g_free (user->gecos);
 
-       g_clear_pointer (&user->login_history, g_variant_unref);
+        g_clear_pointer (&user->login_history, g_variant_unref);
+        g_clear_pointer (&user->user_expiration_time, g_date_time_unref);
+        g_clear_pointer (&user->last_change_time, g_date_time_unref);
 
         if (G_OBJECT_CLASS (user_parent_class)->finalize)
                 (*G_OBJECT_CLASS (user_parent_class)->finalize) (object);
